"""Representation of the contents of a cluster.

A manifest may be built directly from the local context of a cluster, or may be
serialized and stored and checked into the cluster for use in other applications
e.g. such as writing management plan for resources.
"""

import datetime
from dataclasses import dataclass
from typing import Any

from pydantic import BaseModel

__all__ = [
    "Manifest",
    "Cluster",
    "Kustomization",
    "HelmRepository",
    "HelmRelease",
    "HelmChart",
]


@dataclass
class HelmChart:
    """A representation of an instantiation of a chart for a HelmRelease."""

    name: str
    """The name of the chart within the HelmRepository."""

    version: str
    """The version of the chart."""

    repo_name: str
    """The name of the HelmRepository."""

    repo_namespace: str
    """The namespace of the HelmRepository."""

    @classmethod
    def from_doc(cls, doc: dict[str, Any]) -> "HelmChart":
        """Parse a HelmChart from a HelmRelease resource object."""
        if not (spec := doc.get("spec")):
            raise ValueError(f"Invalid {cls} missing spec: {doc}")
        if not (chart := spec.get("chart")):
            raise ValueError(f"Invalid {cls} missing spec.chart: {doc}")
        if not (chart_spec := chart.get("spec")):
            raise ValueError(f"Invalid {cls} missing spec.chart.spec: {doc}")
        if not (chart := chart_spec.get("chart")):
            raise ValueError(f"Invalid {cls} missing spec.chart.spec.chart: {doc}")
        if not (version := chart_spec.get("version")):
            raise ValueError(f"Invalid {cls} missing spec.chart.spec.version: {doc}")
        if not (source_ref := chart_spec.get("sourceRef")):
            raise ValueError(f"Invalid {cls} missing spec.chart.spec.sourceRef: {doc}")
        if "namespace" not in source_ref or "name" not in source_ref:
            raise ValueError(f"Invalid {cls} missing sourceRef fields: {doc}")
        return cls(chart, version, source_ref["name"], source_ref["namespace"])

    @property
    def chart_name(self) -> str:
        """Identifier for the HelmChart."""
        return f"{self.repo_namespace}-{self.repo_name}/{self.name}"


@dataclass
class HelmRelease:
    """A representation of a Flux HelmRelease."""

    name: str
    """The name of the HelmRelease."""

    namespace: str
    """The namespace that owns the HelmRelease."""

    chart: HelmChart
    """A mapping to a specific helm chart for this HelmRelease."""

    @classmethod
    def from_doc(cls, doc: dict[str, Any]) -> "HelmRelease":
        """Parse a HelmRelease from a kubernetes resource object."""
        if not (metadata := doc.get("metadata")):
            raise ValueError(f"Invalid {cls} missing metadata: {doc}")
        if not (name := metadata.get("name")):
            raise ValueError(f"Invalid {cls} missing metadata.name: {doc}")
        if not (namespace := metadata.get("namespace")):
            raise ValueError(f"Invalid {cls} missing metadata.namespace: {doc}")
        chart = HelmChart.from_doc(doc)
        return cls(name, namespace, chart)

    @property
    def release_name(self) -> str:
        """Identifier for the HelmRelease."""
        return f"{self.namespace}-{self.name}"


@dataclass
class HelmRepository:
    """A representation of a flux HelmRepository."""

    name: str
    """The name of the HelmRepository."""

    namespace: str
    """The namespace of owning the HelmRepository."""

    url: str
    """The URL to the repository of helm charts."""

    @classmethod
    def from_doc(cls, doc: dict[str, Any]) -> "HelmRepository":
        """Parse a HelmRepository from a kubernetes resource."""
        if not (metadata := doc.get("metadata")):
            raise ValueError(f"Invalid {cls} missing metadata: {doc}")
        if not (name := metadata.get("name")):
            raise ValueError(f"Invalid {cls} missing metadata.name: {doc}")
        if not (namespace := metadata.get("namespace")):
            raise ValueError(f"Invalid {cls} missing metadata.namespace: {doc}")
        if not (spec := doc.get("spec")):
            raise ValueError(f"Invalid {cls} missing spec: {doc}")
        if not (url := spec.get("url")):
            raise ValueError(f"Invalid {cls} missing spec.url: {doc}")
        return cls(name, namespace, url)

    @property
    def repo_name(self) -> str:
        """Identifier for the HelmRepository."""
        return f"{self.namespace}-{self.name}"


@dataclass
class Kustomization:
    """A Kustomization is a set of declared cluster artifacts.

    This represents a flux Kustomization that points to a path that
    contains typical `kustomize` Kustomizations on local disk that
    may be flat or contain overlays.
    """

    name: str
    """The name of the kustomization."""

    path: str
    """The local repo path to the kustomization."""

    helm_repos: list[HelmRepository]
    """The set of HelmRepositories represented in this kustomization."""

    helm_releases: list[HelmRelease]
    """The set of HelmRelease represented in this kustomization."""

    @property
    def id_name(self) -> str:
        """Identifier for the Kustomization in tests"""
        return f"{self.path}"


@dataclass
class Cluster:
    """A set of nodes that run containerized applications.

    Many flux git repos will only have a single flux cluster, though
    a repo may also contain multiple (e.g. dev an prod).
    """

    name: str
    """The name of the cluster."""

    path: str
    """The local git repo path to the Kustomization objects for the cluster."""

    kustomizations: list[Kustomization]
    """A list of flux Kustomizations for the cluster."""

    def helm_repo_config(self) -> dict[str, Any]:
        """Return a synthetic HelmRepoistory config."""
        now = datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)
        repos = []
        for kustomize in self.kustomizations:
            repos.extend(
                [
                    {
                        "name": f"{repo.namespace}-{repo.name}",
                        "url": repo.url,
                    }
                    for repo in kustomize.helm_repos
                ]
            )
        return {
            "apiVersion": "",
            "generated": now.isoformat(),
            "repositories": repos,
        }

    @property
    def id_name(self) -> str:
        """Identifier for the Cluster in tests."""
        return f"{self.path}"


class Manifest(BaseModel):
    """Holds information about cluster and applications contained in a repo."""

    clusters: list[Cluster]
    """A list of Clusters represented in the repo."""



"""Library for generating kustomize commands to build local cluster resources.

Kustomize build can be used to apply overlays and output a set of resources or artifacts
for the cluster that can be either be parsed directly or piped into additional
commands for processing and filtering by kustomize grep.

This example returns the objects inside a Kustomization using `kustomize build`:
```python
from flux_local.kustomize import Kustomize

objects = await Kustomize.build('/path/to/objects').objects()
for object in objects:
    print(f"Found object {object['apiVersion']} {object['kind']}")
```

You can also filter documents to specific resource types or other fields:
```python
from flux_local.kustomize import Kustomize

objects = await Kustomize.build('/path/to/objects').grep('kind=ConfigMap').objects()
for object in objects:
    print(f"Found ConfigMap: {object['metadata']['name']}")
```

You can apply kyverno policies to the objects with the `validate` method.
"""

from pathlib import Path
from typing import Any, AsyncGenerator

import yaml

from . import command

KUSTOMIZE_BIN = "kustomize"
KYVERNO_BIN = "kyverno"


class Kustomize:
    """Library for issuing a kustomize command."""

    def __init__(self, cmds: list[list[str]]) -> None:
        """Initialize Kustomize."""
        self._cmds = cmds

    @classmethod
    def build(cls, path: Path) -> "Kustomize":
        """Build cluster artifacts from the specified path."""
        return Kustomize(cmds=[[KUSTOMIZE_BIN, "build", str(path)]])

    def grep(
        self, expr: str, path: Path | None = None, invert: bool = False
    ) -> "Kustomize":
        """Filter resources based on an expression.

        Example expressions:
          `kind=HelmRelease`
          `metadata.name=redis`
        """
        out = [KUSTOMIZE_BIN, "cfg", "grep", expr]
        if invert:
            out.append("--invert-match")
        if path:
            out.append(str(path))
        return Kustomize(self._cmds + [out])

    async def run(self) -> str:
        """Run the kustomize command and return the output as a string."""
        return await command.run_piped(self._cmds)

    async def _docs(self) -> AsyncGenerator[dict[str, Any], None]:
        """Run the kustomize command and return the result documents."""
        out = await self.run()
        for doc in yaml.safe_load_all(out):
            yield doc

    async def objects(self) -> list[dict[str, Any]]:
        """Run the kustomize command and return the result cluster objects as a list."""
        return [doc async for doc in self._docs()]

    async def validate(self, policy_path: Path) -> None:
        """Apply kyverno policies from the directory to any objects built so far.

        The specified `policy_path` is a file or directory containing policy objects. All
        secrets will stripped since otherwise they fail the kyverno cli.
        """
        kustomize = self.grep("kind=^Secret$", invert=True)
        cmds = kustomize._cmds + [  # pylint: disable=protected-access
            [
                KYVERNO_BIN,
                "apply",
                str(policy_path),
                "--resource",
                "-",
            ],
        ]
        await command.run_piped(cmds)



"""Library for issuing commands using asyncio and returning the result."""

import asyncio
import logging
import shlex
import subprocess

_LOGGER = logging.getLogger(__name__)


class CommandException(Exception):
    """Error while running a command."""


async def run_piped(cmds: list[list[str]]) -> str:
    """Run a set of commands, piped together, returning stdout of last."""
    stdin = None
    out = None
    for cmd in cmds:
        cmd_text = " ".join([shlex.quote(arg) for arg in cmd])
        _LOGGER.debug("Running command: %s", cmd_text)
        proc = await asyncio.create_subprocess_shell(
            cmd_text,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        out, err = await proc.communicate(stdin)
        if proc.returncode:
            if out:
                _LOGGER.error(out.decode("utf-8"))
            if err:
                _LOGGER.error(err.decode("utf-8"))
            raise CommandException(
                f"Subprocess '{cmd_text}' failed with return code {proc.returncode}: "
                + out.decode("utf-8")
            )
        stdin = out
    return out.decode("utf-8") if out else ""


async def run(cmd: list[str]) -> str:
    """Run the specified command and return stdout."""
    return await run_piped([cmd])
